"""
Module with classes to help with the specification and management of and
interaction with data annotations. Annotations are used to describe
specific subsets of data in further detail. This is used, e.g., to
describe the location of data channels or to mark events in time.

"""
import warnings
import numpy as np
import h5py

from brain.dataformat.base import ManagedGroup, ManagedObject


class DataSelection(object):
    """
    A single data data_selection for a given dataset.

    **Basic Structure**

    The data_selection object defines a set of selections (one per axis).
    Axes for which no data_selection is specified are interpreted as an ALL data_selection,
    i.e., [1,1,1,...,1] is assumed. The individual selections are assumed to be
    combined via a binary AND operation, i.e., only objects selected by all
    selections are seen as being selected. A data_selection may also be applied to the
    -1 axis, indicating that the data_selection is applied to all axes. In the case
    of such a global data_selection, the data_selection is described via a bool matrix
    of the same shape as the data (rather than a vector). As such, specifying a
    global (-1) data_selection is generally something that should be avoided if possible,
    because this may result in very large data_selection arrays, as a bool value must be
    created for each data element.

    **Creating and Updating Data Selections**

    A default DataSelection my be created simply by specifying only the data object the
    selection applies to without specifying any explicit selection:

    >>> data = np.arange(100).reshape((10, 10))
    >>> d1 = DataSelection(data_object=data)

    This will create a default DataSelection with an empty selection dict, i.e.,
    `d1.selections = {}`. As missing selections are treated as ALL, this results in
    a selection data implicitly selects all objects.

    We may now change the selection, simply using array slicing. In the slicing we
    specify the axis index first, and then the elements to be changed:

    >>> d1[0, 1:3] = True  # axis=0, selection=1:3
    >>> print d1[0]
    [ True  True  True False False False False False False False]

    *NOTE:* While missing selections are treated as ALL in general, if a selection is
    missing during assignment, we will initialize the missing selection as all False.
    This divergence from the standard scheme achieves that all values outside of the
    given first assignment are not selected . This divergence is made to:
    i) allow the more intuitive usage for users of selecting elements of interest
    via assignment rather then having to define all elements that are not of
    interest and ii) to avoid the problem that elements are automatically selected
    for the user.

    *NOTE:* Instead of specifying the index of the axis, we may also use the label
    of the dimension if the data object the DataSelection applies is an h5py.Dataset
    with dimension scales. E.g, `d1['time', 1:3] = True`

    Instead of creating a default selection and then updating it, we can also directly
    create a custom selection, by directly initializing the selections dictionary
    during creation of the DataSelection object:

    >>>t = np.zeros(10, 'bool')
    >>>t[0:3] = True
    >>>d1 = DataSelection(data_object=data, selections= {0: t})

    **Accessing Axis Selection**

    To access the selections associated with the different axes we can do
    one of the following two things:

    >>> print d1[0]  # Get selection for axis 0 using array slicing
    [ True  True  True False False False False False False False]
    >>> print d1.selections[0]  # Get selection for axis 0 by accessing the selection dict directly
    [ True  True  True False False False False False False False]

    As shown above, we can retrieve the selection associated using standard slicing
    against the DataSelection directly or by slicing into the selections dict. For
    axis with an explicit selection, both methods result in the same behavior.
    However, for axes without a selection, the first approch will yield None, while
    the latter approach will result in a KeyError:

    >>> print d1[1]
    >>> print d1.selections[1]
    Traceback (most recent call last):
    File "<stdin>", line 1, in <module>
    KeyError: 1

    **Logical Operations / Combining Data Selection**

    DataSelection objects may be combined via bitwise binary operators:

        * `&` : AND operator defining the intersection of two DataSelections `s1 & s2`
        * `|` : OR operator defining the merge of two DataSelections `s1 | s2`
        * `^` : XOR of two data DataSelections `s1 ^ s2`
        * `~` : NOT operator inverting a DataSelection `~s1`

    **Comparing Data Selections**

        * `in` : `s1 in s2' Check if DataSelection s1 is a subset of s2 `s1 in s2`
        * '>=` : `s1 >= s2` : Check if the DataSelection s1 contains all elements of s2.
        * '>' : `s1 > s`' : Check if the DataSelection s1 contains all elements of s2 and has additional elements
        * '==' : `s1 == s2` : Check if the DataSelection s1 selects exactly the same elements as s2
        * '!=' : `s1 != s2` : Check if the DataSelection s1 selects at least in parts different elements than s2
        * '<' : `s1 < s2` : Check if the DataSelection s1 is contained in s2 but s2 selects additional elements as well
        * '<=' : `s1 <= s2` : Check if the DataSelection s1 is a subset of s2
        * `>>` : `s1 >> s2` : Check if s1 follows s2, i.e., all elements selected by s1 appear after all elements \
                              selected by s2.
        * `<<` : `s1 >> s2` : Check if s1 precedes s2, i.e., all elements selected by s1 appear before all elements \
                              selected by s2.

    **Other Selection Operations**

        * `validate()` : Validate that a data_selection is well-defined/valid
        * `len` : The python len operator functions is equivalent to calling count() (see next).
        * `count()` : Count the total number of elements selected. This is equivalent to `len(a)`
        * `simplify_selection()` : Remove unnecessary components of the data_selection.
        * `collapse_selection()` : Collapse the data_selection to a single global data_selection.

    :ivar selections: Dict of numpy bool vector (True/False) indicating which values are selected
                      for a given axis. The axis is the key and data_selection vectors are the values.
                      Use -1 for a complex data_selection that is applied to all axes. As soon as a
                      a global -1 data_selection is present could also be collapsed to a single
                      global data_selection.

    :ivar data_object: The object to which the data_selection is applied to. The object must support
                      the .shape attribute and support h5py/numpy array slicing.
    """
    def __init__(self, data_object, selections=None, simplify=False, collapse=False):
        """
        Initialize a new DataSelection object.

        :param selections: Dict of numpy bool vector (True/False) indicating which values are selected
                      for a given axis. The axis is the key and data_selection vectors are the values.
                      Use -1 for a complex data_selection that is applied to all axes. As soon as a
                      a global -1 data_selection is present could also be collapsed to a single
                      global data_selection. Use None or an empty dict {} to create a basic selection
                      of the whole data.
        :param data_object: The object to which the data_selection is applied to. The object must support
                      the .shape attribute and support h5py/numpy array slicing.
        :param simplify: Simplify the incoming data_selection if possible (see also simplify_selection(..))
        :param collapse: Collapse the data_selection if possible (see also collapse_selection(..) with global_only=True)

        """
        # Save the data_selection properties
        if selections is not None:
            self.selections = selections
        else:
            self.selections = {}
        self.data_object = data_object
        # Simplify and collapse the data_selection if requested
        if simplify:
            self.simplify_selection()
        if collapse:
            self.collapse_selection(global_only=True)

    def __setitem__(self, key, value):
        """
        Define the selection for a given axis using the following syntax.

        s[axis_index, value_selection] = True

        The right-hand assignment value must be a bool or bool-array. The axis_index
        must be an integer indicating a valid axis. The selection may be any
        valid data selection along the given axis.

        Note, if not selection exists yet for the given axis, then all values
        outside of the given assignment will be set to False. This is in contrast
        to the fact that non-existent selections are generally treated as all 1.
        This divergence is made to: i) allow the more intuitive usage for users of
        selecting elements of interest via assignment rather then having to define
        all elements that are not of interest and ii) avoid the problem that
        elements are automatically selected for the user.
        """
        # Check if the given key/value are valid
        if not isinstance(key, tuple) or len(key) != 2:
            raise KeyError("Invalid key for assignment to DataSelection. Expected tuple of (axis_index, selection)")
        axis_index = key[0]
        axis_selection = key[1]
        # Look up the axis if it is given by name
        if isinstance(axis_index, basestring):
            try:
                dim_labels = [dim.label for dim in self.data_object.dims]
            except AttributeError:
                dim_labels = []
            if axis_index in dim_labels:
                axis_index = dim_labels.index(axis_index)

        if not isinstance(axis_index, int):
            raise KeyError("Axis index must be an integer.")
        num_dims = len(self.data_object.shape)
        if axis_index < -1 or axis_index >= num_dims:
            raise KeyError("Axis index out-of-bounds.")

        #Check if the given axis already has a selection
        if not axis_index in self.selections:
            if axis_index >= 0:
                new_selection = np.zeros(shape=(self.data_object.shape[axis_index],), dtype='bool')
            else:  # axis_index == -1
                new_selection = np.zeros(shape=self.data_object.shape, dtype='bool')
            new_selection[axis_selection] = value
            self.selections[axis_index] = new_selection
        else:
            # Assign the given selection
            self.selections[axis_index][axis_selection] = value

    def __getitem__(self, item):
        """
        Get the selection for the given axis. If no selection is present
        for the specified axis, then None is returned instead
        """
        if item in self.selections:
            return self.selections[item]
        elif hasattr(self.data_object, 'dims'):
            try:
                dim_labels = [dim.label for dim in self.data_object.dims]
                if item in dim_labels:
                    dim_index = dim_labels.index(item)
                    return self.selections[dim_index]
                else:
                    return None
            except AttributeError:
                return None
        else:
            return None

    def __and__(self, other):
        """
        Return a new DataSelection that defines a logical AND of the two selections.
        Selections are combined on a per-axis basis.

        NOTE: The new data_selection will always be defined on the data_object of the left-hand
              DataSelection A of (A and B).

        NOTE: If only one of the two selections defines a data_selection for a given axes
        then the single data_selection that is defined will be used, implicitly interpreting
        the missing data_selection as an ALL [1,1,1....1] data_selection.

        :param other: The right-hand DataSelection object B of the A & B operation.
        :type other: DataSelection

        """
        if isinstance(other, DataSelection):
            # Check if the shapes of the data objects match
            if self.data_object.shape != other.data_object.shape:
                raise AttributeError("Axis mismatch between data objects the data_selection is applied to.")
            if not self.validate() or not other.validate():
                raise ValueError("Invalid data_selection.")
            all_axes = set(self.axes() + other.axes())  # Get all axes that have explicit selections
            new_selection = {}  # Create empty new data_selection
            # Iterate over all selections
            for axis in all_axes:
                # Merge the two selections if both objects define a data_selection for the axis
                if axis in self.axes() and axis in other.axes():
                    new_selection[axis] = self.selections[axis] & other.selections[axis]
                # Keep the left-hand data_selection if only it defines an explicit data_selection for the axis
                elif axis in self.axes():
                    new_selection[axis] = self.selections[axis]
                # Keep the right-hand data_selection if only it defines an explicit data_selection for the axis
                elif axis in other.axes():
                    new_selection[axis] = other.selections[axis]
            # Create a new DataSelection object and simplify and collapse the data_selection if possible
            return DataSelection(selections=new_selection,
                                 data_object=self.data_object,
                                 simplify=True,
                                 collapse=True)
        else:
            raise ValueError('AND operations not supported for objects other than DataSelections')

    def __or__(self, other):
        """
        Return a new DataSelection that defines the logical OR of the two selections.

        NOTE: The new data_selection will always be defined on the data_object of the left-hand
              DataSelection A of (A or B).

        NOTE: Due to the complexity of the resulting data_selection, the returned DataSelection
        may define a complex, global data_selection for axis -1. Only in the case where
        both selections define just a single data_selection on the same axis do we keep
        a simple data_selection. Global selections may require large memory as we need
        to store one bool for each element in the dataset.

        :param other: The right-hand DataSelection object B of the A & B operation.
        :type other: DataSelection

        """
        if isinstance(other, DataSelection):
            # Check if the shapes of the data objects match
            if self.data_object.shape != other.data_object.shape:
                raise AttributeError("Axis mismatch between data objects the data_selection is applied to.")
            if not self.validate() or not other.validate():
                raise ValueError("Invalid data_selection.")
            all_axes = list(set(self.axes() + other.axes()))  # Get all axes with an explicit data_selection
            new_selection = {}
            if -1 in all_axes or len(all_axes) > 1:  # If at least one of the selections defines a global data_selection
                # Create new collapsed selections and merge them with or
                left_hand = self.__collapse_selection__(selection_dict=self.selections,
                                                        data_shape=self.data_object.shape,
                                                        global_only=False)[-1]
                right_hand = self.__collapse_selection__(selection_dict=other.selections,
                                                         data_shape=self.data_object.shape,
                                                         global_only=False)[-1]
                new_selection = {-1: left_hand | right_hand}
            elif len(all_axes) == 0:  # Both selections select all data
                new_selection = {}  # Select everything
            elif len(all_axes) == 1:  # One or both selections define a data_selection on a single dimension
                axis = all_axes[0]
                if axis in self.axes() and axis in other.axes():
                    # Create a new data_selection where the two selections are merged
                    new_selection = {axis: self.selections[axis] | other.selections[axis]}
                else:
                    # One defines a data_selection for the axis and the other selects everything
                    new_selection = {}  # Select everything
            return DataSelection(selections=new_selection,
                                 data_object=self.data_object,
                                 simplify=False,
                                 collapse=False)
        else:
            raise ValueError('OR operations not supported for object other than DataSelections')

    def __xor__(self, other):
        """
        Return a new data data_selection that defines an XOR between the two selections.
        NOTE: If an data_selection does not define a data_selection for an axis, then it is
        interpreted as having an ALL data_selection, i.e., [1,1,1,...1]. As such for
        axes for which only one side defines an explicit data_selection, the operation
        equivalent ot an invert/negation of the data_selection that exists.
        NOTE: Due to the complexity of the XOR this operation typically result in a global
        -1 axis data_selection in order to be able to represent the resulting data_selection,
        even if the inputs are all just per-axis selections. Only in the case where
        both selections define just a single data_selection on the same axis do we keep
        a simple data_selection. Global selections may require large memory as we need
        to store one bool for each element in the dataset.
        """
        if isinstance(other, DataSelection):
            # Check that the data_selection shapes match
            if self.data_object.shape != other.data_object.shape:
                raise AttributeError("Axis mismatch between data objects the data_selection is applied to.")
            if not self.validate() or not other.validate():
                raise ValueError("Invalid data_selection.")
            # Get all axes for which we have selections
            all_axes = list(set(self.axes() + other.axes()))
            if (len(all_axes) > 1) or (-1 in all_axes):
                left_hand = self.__collapse_selection__(selection_dict=self.selections,
                                                        data_shape=self.data_object.shape,
                                                        global_only=False)[-1]
                right_hand = self.__collapse_selection__(selection_dict=other.selections,
                                                         data_shape=self.data_object.shape,
                                                         global_only=False)[-1]
                new_selection = {-1: left_hand ^ right_hand}
            elif len(all_axes) == 1:
                new_selection = {}
                axis = all_axes[0]
                if axis in self.axes() and axis in other.axes():
                    new_selection[axis] = self.selections[axis] ^ other.selections[axis]
                elif axis in self.axes():
                    new_selection[axis] = ~self.selections[axis]
                elif axis in other.axes():
                    new_selection[axis] = ~other.selections[axis]
            else:  # i.e. we do not have any explicit data_selection specified.
                # Create a new data_selection that selects nothing
                data_shape = self.data_object.shape
                num_axes = len(data_shape)
                new_selection = {axis: np.zeros(data_shape[axis], dtype='bool') for axis in range(len(num_axes))}
            return DataSelection(selections=new_selection,
                                 data_object=self.data_object,
                                 simplify=False,
                                 collapse=False)
        else:
            raise ValueError('XOR operations not supported for object other than DataSelections')

    def __rshift__(self, other):
        """
        `a >> b` : Identify if a follows b.

        This checks for all selections applied whether all elements selected by `a` appear
        after the elements selected by `b`. This means both `a >> b` and `b >> a` can be
        False at the same time in the case that either a and b overlap in their selection ranges.
        """
        lower_bounds_self, upper_bounds_self = self.__axis_bounds__()
        lower_bounds_other, upper_bounds_other = other.__axis_bounds__()
        all_axes = list(set(self.axes(restricted_only=True, bounds=(lower_bounds_self, upper_bounds_self)) +
                            other.axes(restricted_only=True, bounds=(lower_bounds_other, upper_bounds_other))))
        return np.all(lower_bounds_self[all_axes] > upper_bounds_other[all_axes])

    def __lshift__(self, other):
        """
        `a << b` : Identify if a precedes b.

        This checks for all selections applied whether all elements selected by `a` appear
        before the elements selected by `b`. This means both `a >> b` and `b >> a` can be
        False at the same time in the case that either a and b overlap in their selection ranges.
        """
        lower_bounds_self, upper_bounds_self = self.__axis_bounds__()
        lower_bounds_other, upper_bounds_other = other.__axis_bounds__()
        all_axes = list(set(self.axes(restricted_only=True, bounds=(lower_bounds_self, upper_bounds_self)) +
                            other.axes(restricted_only=True, bounds=(lower_bounds_other, upper_bounds_other))))
        return np.all(upper_bounds_self[all_axes] < lower_bounds_other[all_axes])

    def __invert__(self):
        """
        Return a new DataSelection object that inverts the current data_selection.

        NOTE: This may result in a complex, global data_selection. Only in the case where we have no or only
        have a single data_selection applied to one axes can we maintain a simple per-axis data_selection.
        Global selections may require large memory as we need to store
        one bool for each element in the dataset.

        NOTE: The data_selection of this data_selection will be simplified first in an attempt to reduce
        the number of axes selections used with the goal to keep a simple per-axis data_selection
        object if possible.
        """
        if not self.validate():
            raise ValueError("Invalid data_selection.")
        self.simplify_selection()
        num_axes = len(self.axes())
        new_selection = {}
        if num_axes == 0:
            # Define a new data_selection that selects no objects
            data_shape = self.data_object.shape
            new_selection = {axis: np.zeros(data_shape[axis], dtype='bool') for axis in range(len(num_axes))}
        if num_axes == 1:
            new_selection = {k: ~v for k, v in self.selections.iteritems()}  # The for loop is only 1 iteration
        else:
            # Define a new global data_selection that inverts the current data_selection
            new_selection.update(self.__collapse_selection__(selection_dict=self.selections,
                                                             data_shape=self.data_object.shape,
                                                             global_only=False))
            new_selection[-1] = ~(new_selection[-1])
        return DataSelection(selections=new_selection,
                             data_object=self.data_object,
                             simplify=False,
                             collapse=False)

    def __lt__(self, other):
        """
        Less than operator (<)
        """
        return self.__comparison_operator__(other=other,
                                            operator='lt')

    def __le__(self, other):
        """
        Less than or equals operator (<=)
        """
        return self.__comparison_operator__(other=other,
                                            operator='le')

    def __eq__(self, other):
        """
        Equals operator (==)
        """
        return self.__comparison_operator__(other=other,
                                            operator='eq')

    def __ne__(self, other):
        """
        Not equals operator (!=)
        """
        return self.__comparison_operator__(other=other,
                                            operator='ne')

    def __gt__(self, other):
        """
        Greater than operator (>)
        """
        return self.__comparison_operator__(other=other,
                                            operator='gt')

    def __ge__(self, other):
        """
        Greater than or equals operator (>=)
        """
        return self.__comparison_operator__(other=other,
                                            operator='ge')

    def __contains__(self, other):
        """
        Check whether a data_selection is a subset of the other data_selection.
        """
        return self.__comparison_operator__(other=other,
                                            operator='in')

    def __len__(self):
        """
        Get number of selected elements.
        """
        return self.count()

    def __axis_bounds__(self):
        """
        Compute the smallest and largest selected index along all axes.

        :returns: This function returns a tuple of two lists, one with the lower axis bounds and one with the
                  upper axis bounds.
        """
        lower_axis_bounds = np.zeros(shape=(len(self.data_object.shape), ), dtype='int64')  # Init lower bounds as 0
        upper_axis_bounds = np.asarray(self.data_object.shape) - 1  # Init the upper bounds as max index
        for axis, select in self.selections.iteritems():
            if axis >= 0:
                lower_axis_bounds[axis] = np.argmax(select)
                upper_axis_bounds[axis] = select.size - np.argmax(select[::-1])
        # If we have a global selection we need to compute the bounds for it and compare
        if -1 in self.axes():
            for axis in range(len(self.data_object.shape)):
                collapse_axis = tuple([i for i in range(len(self.data_object.shape)) if i != axis])
                numpy_version = np.version.version.split('.')  # Check numpy versions
                # Let numpy handle the collapse of the matrix if possible (numpy version >= 1.7 does this)
                if int(numpy_version[0]) > 1 or (int(numpy_version[0]) == 1 and int(numpy_version[1]) >= 7):
                    collapsed_selection = np.any(self.selections[-1], axis=tuple(collapse_axis))
                else:  # If the numpy version is too old then we'll do it ourselves
                    collapsed_selection = self.selections[-1]
                    for caxis in reversed(collapse_axis):
                        collapsed_selection = np.max(collapsed_selection, axis=caxis)
                current_lower_bound = np.argmax(collapsed_selection)
                current_upper_bound = collapsed_selection.size - np.argmax(collapsed_selection[::-1])
                if current_lower_bound > lower_axis_bounds[axis]:
                    lower_axis_bounds[axis] = current_lower_bound
                if current_upper_bound < upper_axis_bounds[axis]:
                    upper_axis_bounds[axis] = current_upper_bound
        return lower_axis_bounds, upper_axis_bounds

    def __comparison_operator__(self, other, operator):
        """
        Internal helper function used to implement a series of comparison operators.

        NOTE: These operations checks are not just comparisons of one selection selecting
              more or less items then the other, but are true overlap containment checks.
              E.g., a<b means that b must select all times that a selects and b must
              select additional items that a does not select. This also means that as
              soon as a selects a record that is not in b, a<b will be False.

        :param other: The other DataSelection object
        :type other: DataSelection
        :param operator: String indicating the comparison operator to be used, which include:
                         'in', 'ge' is >= , 'gt' is >, 'eq' is ==, 'ne is !=, 'le' is <=, 'lt' is <.
        :type: str or unicode

        """
        # Check if we can compare agains the give object
        if isinstance(other, DataSelection):
            # Not equal ne is implemented by inverting the result of equals check
            if operator == 'ne':
                return not self.__comparison_operator__(other=other, operator='eq')

            # If we have a global selection in any of the two objects then we need to
            # create global selections for both of two and compare them directly
            if -1 in other.axes() or -1 in self.axes():
                my_selection = self.__collapse_selection__(selection_dict=self.selections,
                                                           data_shape=self.data_object.shape,
                                                           global_only=False)[-1]
                other_selection = self.__collapse_selection__(selection_dict=other.selections,
                                                              data_shape=self.data_object.shape,
                                                              global_only=False)[-1]
                my_count = my_selection.sum()
                other_count = other_selection.sum()
                combined_count = (my_selection & other_selection).sum()
                if operator == 'in' or operator == 'ge':
                    return combined_count == other_count
                elif operator == 'lt':
                    return (combined_count == my_count) and (my_count < other_count)
                elif operator == 'gt':
                    return (combined_count == other_count) and (my_count > other_count)
                elif operator == 'eq':
                    return (combined_count == other_count) and (my_count == other_count)
                elif operator == 'ne':  # This is handled separetely above
                    pass  # return not ((combined_count==other_count) and (my_count==other_count))
                elif operator == 'le':
                    return combined_count == my_count
                else:
                    raise ValueError("Unsupported comparison operator given.")
            # If we only have per-axis selections defined then we can check this by comparing
            # the selections individually without creating a large global seletion
            else:
                # 1) Compute all axes used
                all_axes = list(set(self.axes() + other.axes()))

                # 2) If no selections are specified, ie., both are selecting everything, hence both are equal
                if len(all_axes) == 0:
                    return operator in ['eq', 'le', 'ge', 'in']

                # 2) Compute the result by comparing the selection of all axes
                results = np.zeros(len(all_axes), dtype='bool')
                logical_result = True
                for index, axis in enumerate(all_axes):
                    # 2.1) Contains and greater equal
                    if operator == 'in' or operator == 'ge':
                        if axis not in self.axes():
                            continue
                        elif axis not in other.axes():
                            if self.selections[axis].sum() < self.data_object.shape[axis]:
                                logical_result = False
                                break
                        else:
                            other_count = other.selections[axis].sum()
                            post_count = (self.selections[axis] & other.selections[axis]).sum()
                            if post_count != other_count:
                                logical_result = False
                                break
                    # 2.2 Less than
                    elif operator == 'lt':
                        if axis not in self.axes():
                            if other.selections[axis].sum() < self.data_object.shape[axis]:
                                logical_result = False
                                break
                        elif axis not in other.axes():
                            continue
                        else:
                            my_count = self.selections[axis].sum()
                            other_count = other.selections[axis].sum()
                            post_count = (self.selections[axis] & other.selections[axis]).sum()
                            if not ((post_count == my_count) and (my_count < other_count)):
                                logical_result = False
                                break
                    # 2.3 Greater than
                    elif operator == 'gt':
                        if axis not in self.axes():
                            if other.selections[axis].sum() < self.data_object.shape[axis]:
                                results[index] = True
                        elif axis not in other.axes():
                            if self.selections[axis].sum() < self.data_object.shape[axis]:
                                return False
                            else:
                                results[index] = False
                        else:
                            my_count = self.selections[axis].sum()
                            other_count = other.selections[axis].sum()
                            post_count = (self.selections[axis] & other.selections[axis]).sum()
                            if (my_count == other_count) and (post_count == other_count):
                                results[index] = False
                            elif post_count != other_count:
                                return False
                            elif (my_count > other_count) and (post_count == other_count):
                                results[index] = True
                            else:
                                results[index] = False
                    # 2.4 Equals
                    elif operator == 'eq':
                        if axis not in self.axes():
                            if other.selections[axis].sum() != self.data_object.shape[axis]:
                                logical_result = False
                                break
                        elif axis not in other.axes():
                            if self.selections[axis].sum() != self.data_object.shape[axis]:
                                logical_result = False
                                break
                        else:
                            my_count = self.selections[axis].sum()
                            other_count = other.selections[axis].sum()
                            post_count = (self.selections[axis] & other.selections[axis]).sum()
                            if not ((my_count == post_count) and (my_count == other_count)):
                                logical_result = False
                                break
                    # 2.5 Not equals. Nothing to be done here. Implemented above by negating the equals (eq, ==) check
                    elif operator == 'ne':
                        pass
                    # 2.6 Less equals
                    elif operator == 'le':
                        if axis not in self.axes():
                            if other.selections[axis].sum() != self.data_object.shape[axis]:
                                logical_result = False
                                break
                        elif axis not in other.axes():
                            pass
                        else:
                            my_count = self.selections[axis].sum()
                            post_count = (self.selections[axis] & other.selections[axis]).sum()
                            if not post_count == my_count:
                                logical_result = False
                                break
                    else:
                        raise ValueError("Unsupported comparison operator given.")
                if operator == 'gt':
                    logical_result = np.any(results)
                return logical_result
        else:
            raise ValueError('IN operations not supported for objects other than DataSelections')

    @staticmethod
    def __collapse_selection__(selection_dict, data_shape, global_only=True):
        """
        Internal helper function used to collapse a set of selections. A data_selection may be
        collapsed to a single global data_selection if a global data_selection is already present.

        :param selection_dict: DataSelection.selections dictionary describing for all
               axes (keys) the selected objects (values).
        :type selection_dict: dict
        :param data_shape: Tuple with the shape of the data to be selected
        :type data_shape: tuple
        :param global_only: Boolean indicating whether the data_selection should always be
               collapsed to a single global data_selection (False) or only in case that a
               global data_selection is already present (True). Default is True.
        :type global_only: bool
        """
        selected_axes = selection_dict.keys()
        num_axes = len(data_shape)
        if global_only and -1 not in selected_axes:
            return selection_dict
        else:
            if -1 in selected_axes:
                global_selection = selection_dict[-1]
            else:
                global_selection = np.ones(shape=data_shape,
                                           dtype='bool')
            for axis, select in selection_dict.iteritems():
                if axis >= 0:
                    expansion = [None for _ in range(num_axes)]
                    expansion[axis] = slice(None)
                    global_selection *= select[tuple(expansion)]
            return {-1: global_selection}

    def collapse_selection(self, global_only=True):
        """
        Collapse the data_selection to a single global data_selection.

        :param global_only: Boolean indicating whether the data_selection should always be
               collapsed to a single global data_selection (False) or only in case that a
               global data_selection is already present (True). Default is True.
        :type global_only: bool
        """
        self.selections = self.__collapse_selection__(selection_dict=self.selections,
                                                      data_shape=self.data_object.shape,
                                                      global_only=global_only)

    def simplify_selection(self):
        """
        Remove unnecessary entries from the data_selection dict, i.e., entries that
        select all elements for a given axis.
        """
        num_elements = np.asarray(self.data_object.shape).prod(axis=None)
        for axis, select in self.selections.iteritems():
            if axis >= 0:
                if select.sum() == self.data_object.shape[axis]:
                    self.selections.pop(axis, None)
            elif select.sum == num_elements:  # i.e., axis == -1
                self.selections.pop(axis, None)

    def axes(self, restricted_only=False, bounds=None):
        """
        Get the list of axes that are sub-selected by this selections.
        This may be an empty list.

        :param restricted_only: If set to true, then only axes that are actually
                                restricted by the selection are returned. This means
                                axes where we find selected values along the whole
                                axes are not included in the selection. This also
                                means, that -1 is removed from the axes and resolved
                                to identify the actually restricted axes.
        :param bounds: Optional input used to indicate the lower and upper bounds
                       selected along a given axis. If set to None, then the bounds
                       will be computed using self.__axis_bounds__.
        :type bounds: Tuple of two lists indicating the lower and upper bounds. Same as
                      output of self.__axis_bounds__

        :returns: List of integers with the axes indices.
        """
        if restricted_only:
            all_axes = []
            for axis, selection in self.selections.iteritems():
                if axis >= 0 and not np.all(selection):
                    all_axes.append(axis)
                elif axis == -1:
                    if bounds is not None:
                        lower_bounds, upper_bounds = bounds
                    else:
                        lower_bounds, upper_bounds = self.__axis_bounds__()
                    for i in range(len(lower_bounds)):
                        if lower_bounds[i] > 0 and upper_bounds[i] < self.data_object.shape[axis]:
                            all_axes.append(i)

                elif axis < -1:
                    KeyError('Invalid axis index smaller -1 in list of selections.')
            return all_axes
        else:
            return self.selections.keys()

    def count(self):
        """
        Get the number of elements selected by the data_selection.
        Dimensions without an explicit data_selection are counted as
        having an ALL data_selection. This is equivalent to calling len() on
        the object.
        """
        axis_counts = self.counts()
        if -1 not in axis_counts:
            counts_array = np.asarray(axis_counts.values())
            return counts_array.prod(axis=None)

        else:
            new_selection = self. __collapse_selection__(selection_dict=self.selections,
                                                         data_shape=self.data_object.shape,
                                                         global_only=False)[-1]
            return new_selection.sum()

    def counts(self):
        """
        Get the number of elements selected by the individual parts of
        the selections defined in the self.selections object.
        Dimensions without an explicit data_selection are interpreted
        as having an ALL data_selection (i.e, the length of that axes
        is returned)

        :returns: Dictionary of {axis_index: count}. Axis -1 indicates
                  the presence of a global data_selection.

        """
        num_selected = {axis: size for axis, size in enumerate(self.data_object.shape)}
        for axis, select in self.selections.iteritems():
            num_selected[axis] = select.sum()
        return num_selected

    def validate(self):
        """
        Check whether the data_selection is valid.
        """
        data_shape = self.data_object.shape
        num_dims = len(data_shape)
        for axis, select in self.selections.iteritems():
            if axis >= num_dims:
                warnings.warn("Axis index out of range.")
                return False
            elif axis >= 0:
                if select.size != data_shape[axis]:
                    warnings.warn("Selection length does not match dimension size.")
                    return False
            elif axis == -1:
                if select.shape != data_shape:
                    warnings.warn("Selection shape does not match data shape.")
                    return False
            else:
                warnings.warn("Invalid axis index in data_selection.")
                return False
        return True

    def data(self):
        """
        Load the data associated with this DataSelection.
        """
        select_axes = self.axes()
        num_axes = len(select_axes)
        if num_axes == 0:  # No selections are specified. Load the full data.
            return self.data_object[:]
        elif -1 in select_axes:  # We have a global selection. Collapse all selection and do one global load.
            if num_axes > 1:
                select = self.__collapse_selection__(selection_dict=self.selections,
                                                     data_shape=self.data_object.shape,
                                                     global_only=False)
            else:
                select = self.selections[-1]
            return self.data_object[select]
        elif num_axes == 1:  # We have single axis selection. We can do this in one data load
            select = [slice(None) for _ in range(len(self.data_object.shape))]
            select[select_axes[0]] = self.selections[select_axes[0]]
            return self.data_object[tuple(select)]
        else:  # We have multiple single axes selections. h5py does not allow us to do this in one load.
            # Sort the data selections by selectivity
            import operator
            axis_counts = self.counts()
            # Compute the number of elements that will be loaded by each selection (if applied alone on the full data)
            selected_elements = {}
            for axis, count in axis_counts.iteritems():
                temp_shape = np.asarray(self.data_object.shape)
                temp_shape[axis] = count
                selected_elements[axis] = np.sum(temp_shape)
            # Sort the selection in increasing order, i.e., perform the data load that loads the fewest elements first
            sorted_axes = sorted(selected_elements.iteritems(), key=operator.itemgetter(1), reverse=False)
            select = [slice(None) for _ in range(len(self.data_object.shape))]
            smallest_axis = sorted_axes.pop(0)[0]
            select[smallest_axis] = self.selections[smallest_axis]
            data = self.data_object[tuple(select)]
            # Now that we have a numpy array (and no longer an h5py object) we can load the remaining data all at once
            select[smallest_axis] = slice(None)
            for axis in sorted_axes:
                select[axis[0]] = self.selections[axis[0]]
            return data[tuple(select)]


class Annotation(object):
    """
    Annotate a particular dataset or subset of data.

    :ivar annotation_type: The type of annotation, e.g., anatomy, event, or general feature.
                           See AnnotationCollection.annotation_type_indexes. Other user-defined types are
                           permitted but it is recommended to chose from the predefined
                           annotation_type_indexes if possible.
    :type annotation_type: str or unicode. Typically one of Annotation.annotation_type_indexes.
    :ivar description: The description of the annotation
    :type description: str or unicode
    :ivar data_selection: The data subset the annotation refers to.
    :type data_selection: DataSelection

    """
    def __init__(self, annotation_type, description, data_selection):
        self.annotation_type = annotation_type
        self.description = description
        self.data_selection = data_selection

    def __logical_operations__(self, other, operation):
        if operation not in ["OR", "AND", "XOR"]:
            raise ValueError("Invalid logical operation specified. " + operation)
        if isinstance(other, Annotation):
            new_description = unicode(self.description) + " " + operation + " " + unicode(other.description)
            if self.annotation_type == other.annotation_type:
                new_annotation_type = self.annotation_type
            else:
                new_annotation_type = self.annotation_type + " " + operation + " " + other.annotation_type
            if operation == "OR":
                new_selection = self.data_selection | other.data_selection
            elif operation == "AND":
                new_selection = self.data_selection & other.data_selection
            elif operation == "XOR":
                new_selection = self.data_selection ^ other.data_selection
            else:
                raise ValueError("Invalid logical operation specified. " + operation)
            return Annotation(annotation_type=new_annotation_type,
                              description=new_description,
                              data_selection=new_selection)
        else:
            raise ValueError(str(operation) + "operations not supported for objects other than Annotations")

    def __or__(self, other):
        return self.__logical_operations__(other=other,
                                           operation='OR')

    def __and__(self, other):
        return self.__logical_operations__(other=other,
                                           operation='AND')

    def __xor__(self, other):
        return self.__logical_operations__(other=other,
                                           operation='XOR')

    def __invert__(self):
        return Annotation(annotation_type=self.annotation_type,
                          description='NOT '+self.description,
                          data_selection=~self.data_selection)

    def __rshift__(self, other):
        """
        Check whether the data selection of the annotation
        follows the selection of the other annotation. Same
        as self.data_selection >> other.data_selection.
        See :func:`DataSelection.__rshift__` for details.
        """
        if isinstance(other, DataSelection):
            return self.data_selection >> other
        elif isinstance(other, Annotation):
            return self.data_selection >> other.data_selection
        else:
            raise ValueError('>> operation not supported for objects other than Annotation and DataAnnotation')

    def __lshift__(self, other):
        """
        Check whether the data selection of the annotation
        precedes the selection of the other annotation. Same
        as self.data_selection >> other.data_selection.
        See :func:`DataSelection.__lshift__` for details.
        """
        if isinstance(other, DataSelection):
            return self.data_selection << other
        elif isinstance(other, Annotation):
            return self.data_selection << other.data_selection
        else:
            raise ValueError('<< operation not supported for objects other than Annotation and DataAnnotation')

    def __lt__(self, other):
        """
        Less than operator (<)
        """
        return self.data_selection < other.data_selection

    def __le__(self, other):
        """
        Less than or equals operator (<=)
        """
        return self.data_selection <= other.data_selection

    def __eq__(self, other):
        """
        Equals operator (==)
        """
        return self.data_selection == other.data_selection

    def __ne__(self, other):
        """
        Not equals operator (!=)
        """
        return self.data_selection != other.data_selection

    def __gt__(self, other):
        """
        Greater than operator (>)
        """
        return self.data_selection > other.data_selection

    def __ge__(self, other):
        """
        Greater than or equals operator (>=)
        """
        return self.data_selection >= other.data_selection

    def __contains__(self, other):
        """
        Check whether a data_selection is a subset of the other data_selection.
        """
        return other.data_selection in self.data_selection

    def __len__(self):
        """
        Get number of selected elements.
        """
        return len(self.data_selection)

    def data(self):
        """
        Load the data associated with this DataSelection.
        """
        return self.data_selection.data()


class AnnotationCollection(object):
    """
    A collection of annotations

    **Instance Variables**

    :ivar data_object: h5py.Dataset, managed object, or numpy array with the data
    :ivar annotation_types:  1D (vlen str) Dataset with the list of all available types (h5py.Dataset or numpy array)
    :ivar annotation_type_indexes: 1D h5py.Dataset or numpy array with the index into the annotation_types array
                                   indicating for each annotation the corresponding annotation type.
    :ivar descriptions: h5py.Dataset or numpy array with the annotation descriptions
    :ivar selection_indexes: h5py.Dataset or numpy array with the indexes of the selections that apply. This is a
                             2D dataset of shape (#annotations, #axes+1). The axes dimension is ordered as [-1,0,1,...].
    :ivar selections:  dict of h5py.Datasets or numpy arrays with all possible selections
    :ivar collection_description: String with a description of the annotation collection

    **Filtering**

    Filtering provides means to locate annotations of interest. Filtering is performed via the
    provided set of filter functions, including:

       * `index_filter(..)` : Select all annotations with the given index.
       * `axis_filter(..)` : Find all annotations that select by a given set of axes
       * `type_filter(..)` : Find all annotations of the given type
       * `description_filer(..)` : Find all annotations with the given description
       * `description_contains_filter(..)` : Find all annotations where the description contains the given string

    Filter functions generate a 1D bool-type numpy array, indicating which annotations are selected by the filter.
    As bool arrays, results of filters may be combined using bitwise logical operations, e.g.,

        * `a & b` : `AND` : Select all annotations selected by both filters `a` and `b`
        * `a | b` : `OR` : Select all annotations that are selected by either filter `a` or `b`
        * `a ^ b` `XOR` : Exclusive or, select all annotations where the filters `a` and `b` differ
        * `~a` : `NOT` : Invert the selection of the filter, selecting all annotations not selected by `a`

    **Selecting:**

    Once we have identified a set of relevant annotations via filtering one can select
    the annotations of interest directly using standard array slicing. E.g,:

    >>> a = AnnotationCollection(..)
    >>> f = a.type_filter('event') & a.axis_filter([0])
    >>> v = a[f]

    The result of selecting annotations is reduced AnnotationCollection object.
    In addition to filter, one can also select annotations directly using
    standard data selection/slicing, e.g, `a[0:10]` to select the first 10 annotations.

    NOTE: When sub-selection is performed, all data relevant to the selected annotations
    will be loaded into memory, whereas when an AnnotationCollection is constructed
    initially it may be initialized using h5py.Dataset objects where the data resides
    in file and is loaded by the filters as needed.

    **Getting the Annotations**

    From the AnnotationCollection we can retrieve a list of all selected Annotation objects
    via the `get_annotations()` functions. This will convert the annotations from
    the collective data structures used for filtering to individual Annotation objects.
    This is typically done after the filtering is complete.

    **Merging Annotations**

    Once we have selected and retrieved the Annotation objects of interest, the
    individual Annotations may be combined using standard bitwise logical operators
    and compared using standard comparison operators (see the documentation of the
    `Annotation` class for further details. As convenience functions, the
    `AnnotationCollection` class provides a set of merge functions which will
    generate a single combined `Annotation` by merging all the selected
    annotations using a given bitwise logical operation, e.g, `merge_and(..)`,
    `merge_or`, and `merge_xor`

    **Other Operations**

        * `len` : Get the number of annotations in the collection using the standartd Python `len(a)`.

    **Example**

    >>> a = AnnotationDataGroup(...)  # Load annotation collection from file
    >>> f = a.type_filter('event') & a.axis_filter([0])  # Find all annotations that define an event on the axis 0
    >>> s = a[f]  # Select all relevant annotations
    >>> s_all = s.get_annotations()  # Get all annotations
    >>> s1 = a.merge_or()  # Define a single combined annotation

    """
    def __init__(self,
                 data,
                 annotation_type_indexes,
                 annotation_types,
                 descriptions,
                 selection_refs,
                 selections,
                 collection_description=''):
        self.data_object = data  # h5py dataset, managed object, or numpy array with the data
        self.annotation_type_indexes = annotation_type_indexes  # hpy dataset/numpy array with the annotation type index
        self.annotation_types = annotation_types  # h5py dataset of numpy array with a list of all available types
        self.descriptions = descriptions  # h5py dataset or numpy array with the annotation descriptions
        self.selection_indexes = selection_refs  # h5py dataset or numpy array with the indexes of the selections used
        self.selections = selections  # dict of h5py datasets or numpy arrays with the selections
        self.collection_description = collection_description  # String with the description of the collections

    def __getitem__(self, selection):
        """
        Get a new AnnotationCollection for the subset of annotations selected.

        NOTE: While most other variables are sub-selects (and loaded into memory) the self.selections
        collection of selections remains unmodified. This strategy i) allows us to keep the selections
        out-of-core in the HDF5 file, and ii) avoids complex updates of references to the selections.
        """
        selected_annotation_type_indexes = self.annotation_type_indexes[selection]
        selected_descriptions = self.descriptions[selection]
        selected_selection_indexes = self.selection_indexes[selection, :]
        selected_selections = self.selections

        return AnnotationCollection(data=self.data_object,
                                    annotation_types=self.annotation_types,
                                    annotation_type_indexes=selected_annotation_type_indexes,
                                    descriptions=selected_descriptions,
                                    selection_refs=selected_selection_indexes,
                                    selections=selected_selections,
                                    collection_description=self.collection_description + " (subset)")

    def __len__(self):
        """
        Get the number of annotations in the collection.
        """
        return len(self.annotation_type_indexes)

    def index_filter(self, selection):
        """
        Filter annotations based on their index.

        :param selection: Any valid selection supported by numpy, e.g., a slice object
               or an integer index. E.g., to select annotation 200 to 300 we could do
               `index_filter(slice(200,301)`. NOTE: Following standard numpy selection
               schema the upper bound is not included in this example.

        :returns: Bool array indicating for each annotation whether it has been selected (True) or not (False)

        """
        relevant_selections = np.zeros((self.num_annotations()), dtype='bool')
        relevant_selections[selection] = True
        return relevant_selections

    def axis_filter(self, axis_list):
        """
        Get all annotations that filter by a given set of axes.

        :param axis_list: List of all axis relevant selections must filter by. Only selections
               that select by all specified axes will be retrieved. The axis may be specified
               by their integer index (recommended, -1,0,1, ... n) or by the name of the
               dimension in the data_object.

        Examples:

        >>> from brain.dataformat.brainformat.braindata import *
        >>> f= BrainDataFile('testfile_real.h5')
        >>> d = f.data().internal().ecog_data(0)
        >>> d.annotations()
        >>> a = d.annotations(0)
        >>> a.axis_filter(0)  # Filter based on a single axis index
        >>> a.axis_filter('space')  # Filter based on a single axis index using the name of the axis
        >>> a.axis_filter(['space', 1])  # Filter based on multiple axis. We here can mix axis names and indexes

        NOTE: Global selections are treated independently and are
        as such if the filter asks for axis=1 selections with
        axis=-1 (global) will not be included.

        :returns: Bool array indicating for each annotation whether it has been selected (True) or not (False)
        """
        # TODO: Add option to allow -1 global selection to be resolved. See also DataSelection.axes -- restricted_only
        relevant_selections = None
        if isinstance(axis_list, int) or isinstance(axis_list, basestring):
            axis_list = [axis_list]
        # Convert any explict axis names to indicies
        try:
            dim_labels = [dim.label for dim in self.data_object.dims]
        except AttributeError:
            dim_labels = []
        if len(dim_labels) > 0:
            for i, axis in enumerate(axis_list):
                if isinstance(axis, basestring):
                    if axis in dim_labels:
                        axis_list[i] = dim_labels.index(axis)
                    else:
                        raise ValueError("Index for axis unknown: " + str(axis))
        else:
            for axis in axis_list:
                if isinstance(axis, basestring):
                    raise ValueError("Index for axis unknown: " + str(axis))

        # Find all selection that select by all given axes
        for axis in axis_list:
            axis_index = axis+1  # Correct the axis index as axes ordered as [-1, 0 , 1, 2, ...]
            if relevant_selections is None:
                relevant_selections = (self.selection_indexes[:, axis_index] >= 0)
            else:
                relevant_selections &= (self.selection_indexes[:, axis_index] >= 0)
        if relevant_selections is None:
            relevant_selections = np.zeros((self.num_annotations()), dtype='bool')
        return relevant_selections

    def type_filter(self, annotation_type):
        """
        Get all annotations with the given type.

        :param annotation_type: Either the string of the annotation type or integer
                                 with the index of the annotation type.

        :returns: Bool array indicating for each annotation whether it has been selected (True) or not (False)
        """
        # If the annotation type is a string then look up the index of the annotation
        if isinstance(annotation_type, basestring):
            find_index = np.where(self.annotation_types[:] == annotation_type)[0]
            if len(find_index) == 0:
                annotation_type_index = -1  # Set to unknown type as all indexes will be positive
            else:
                annotation_type_index = find_index[0]
        else:
            annotation_type_index = annotation_type
        return self.annotation_type_indexes[:] == annotation_type_index

    def type_contains_filter(self, annotation_type):
        """
        Get all annotations where the annotation type contains the given string.

        :param annotation_type: String with annotation type substring to look for

        :returns: Bool array indicating for each annotation whether it has been selected (True) or not (False)
        """
        # In HDF5, data in VL format is stored as arbitrary-length vectors of a base type. In particular, strings
        # are stored C-style in null-terminated buffers. NumPy has no native mechanism to support this.
        # Since numpy does not support this natively, vlen strings are stored as a generic pointer type in the
        # form of the 'object' ('O'). We here need to convert the annotation_type array first to ensure we can
        # apply the char_find methods.
        curr_annotation_types = self.annotation_types[:]
        if self.annotation_types.dtype.kind == 'O':  # We a numpy array of type object
            new_type = '|S' + str(self.annotation_types.dtype.itemsize)
            curr_annotation_types = curr_annotation_types.astype(new_type)

        matching_types = np.where(np.char.find(curr_annotation_types, annotation_type) != -1)[0]
        if matching_types.size == 0:
            return np.zeros(shape=(self.annotation_type_indexes.shape[0],), dtype='bool')
        else:
            matching_annotations = self.annotation_type_indexes[:][..., None] == matching_types[None, ...]
            return np.max(matching_annotations, axis=1)

    def description_filter(self, description):
        """
        Get all annotations with the given description.

        :param description: String of the description to be located.

        :returns: Bool array indicating for each annotation whether it has been selected (True) or not (False)

        """
        return self.descriptions[:] == description

        #curr_descriptions = self.descriptions[:]
        # vlen strings are loaded as object arrays by h5py
        #if self.descriptions.dtype.kind == 'O':  # We have a numpy object type array instead of strings
        #    new_type = '|S' + str(self.descriptions.dtype.itemsize)
        #    curr_descriptions = curr_descriptions.astype(new_type)
        #return np.char.equal(curr_descriptions, description)

    def description_contains_filter(self, description):
        """
        Get all annotations for which the description contains the given text

        :param description: String of the partial description to be found.

        :returns: Bool array indicating for each annotation whether it has been selected (True) or not (False)

        """
        curr_descriptions = self.descriptions[:]
        # vlen strings are loaded as object arrays by h5py
        if self.descriptions.dtype.kind == 'O':  # We have a numpy object type array instead of strings
            new_type = '|S' + str(self.descriptions.dtype.itemsize)
            curr_descriptions = curr_descriptions.astype(new_type)
        return np.char.find(curr_descriptions, description) != -1

    def num_annotations(self):
        """
        Get the number of annotations in the collection.
        """
        return self.annotation_type_indexes.size

    def get_annotations(self):
        """
        Get a list of all annotations.

        :returns: List of Annotation objects

        """
        annotation_list = []
        axis_list = [index for index in range(len(self.data_object.shape))] + [-1]  # List of axes
        # 1) Iterate through all annotations and construct the lit of Annotation objects
        for anno_index in range(self.num_annotations()):
            selection_ref = self.selection_indexes[anno_index, :].reshape(len(axis_list))  # Current selection refs
            anno_type = self.annotation_types[self.annotation_type_indexes[anno_index]]  # Current annotation type
            anno_description = self.descriptions[anno_index]  # Current annotation description
            # Construct the list of selections
            curr_selections = {}
            for axis in axis_list:
                axis_index = axis + 1  # Correct the axis index as axes are stored in oreder of [-1, 0 , 1, 2, ...]
                if selection_ref[axis_index] >= 0:
                    curr_selections[axis] = self.selections[axis][selection_ref[axis_index]]
            # Construct the DataSelection object
            curr_data_selection = DataSelection(data_object=self.data_object,
                                                selections=curr_selections,
                                                simplify=False,
                                                collapse=False)
            # Construct the Annotation object and add it to the list
            annotation_list.append(Annotation(annotation_type=anno_type,
                                              description=anno_description,
                                              data_selection=curr_data_selection))
        # Return the list of annotations
        return annotation_list

    def merge_and(self):
        """
        Get single AnnotationObject (or None) that combines all annotations in this
        collections via logical AND.

        This is a convenience function and is equivalent to calling get_annotations(...)
        and combining all annotations in the returned list via logical & (AND).
        """
        annotation_list = self.get_annotations()
        if len(annotation_list) == 0:
            return None
        else:
            anno = annotation_list[0]
            for anno_index in range(1, len(annotation_list)):
                anno = anno & annotation_list[anno_index]
            return anno

    def merge_or(self):
        """
        Get single AnnotationObject (or None) that combines all annotations in this
        collections via logical OR.

        This is a convenience function and is equivalent to calling get_annotations(...)
        and combining all annotations in the returned list via logical | (OR).
        """
        annotation_list = self.get_annotations()
        if len(annotation_list) == 0:
            return None
        else:
            anno = annotation_list[0]
            for anno_index in range(1, len(annotation_list)):
                anno = anno | annotation_list[anno_index]
            return anno

    def merge_xor(self):
        """
        Get single AnnotationObject (or None) that combines all annotations in this
        collections via logical XOR.

        This is a convenience function and is equivalent to calling get_annotations(...)
        and combining all annotations in the returned list via logical ^ (XOR).
        """
        annotation_list = self.get_annotations()
        if len(annotation_list) == 0:
            return None
        else:
            anno = annotation_list[0]
            for anno_index in range(1, len(annotation_list)):
                anno = anno ^ annotation_list[anno_index]
            return anno

    def add_annotation(self, annotation):
        """
        Add a new annotation to the collection.

        *NOTE:* if the AnnotationCollection object was populated with h5py.Dataset objects,
        e.g., as is the case when using an AnnotationDataGroup instance, then the annotation
        will be written to file. If the AnnotationCollection was initialized as a pure in-memory
        collection using numpy arrays, then the arrays will only be updated in memory.

        *NOTE:* The Annotation is assumed to refer to the same data object as this
        AnnotationCollection. If this is not the case, then the annotation will be either
        reassigned to the data object of the collection (as long as the data object's shape match)
        or a ValueError is raised.

        :param annotation: The annotation to be added
        :type annotation: Annotation

        :raises: ValueError in case the annotation can not be added to the collection
        """
        if not isinstance(annotation, Annotation):
            raise ValueError("The given annotation object is not of type Annotation as expected.")

        data_selection = annotation.data_selection
        new_selections = data_selection.selections
        if self.data_object.shape != data_selection.data_object.shape:
            raise ValueError("The data object of the AnnotationCollection does not " +
                             "match the data object of the new Annotation")

        # Add the selection of the annotation to the selections and keep track of the references
        selection_refs = np.zeros(shape=(self.selection_indexes.shape[1],), dtype='int32') - 1
        axis_list = [-1] + [index for index in range(len(self.data_object.shape))]
        for axis_index, axis in enumerate(axis_list):
            if axis in new_selections:
                axis_selection = self.selections[axis]
                new_shape = list(axis_selection.shape)
                new_shape[0] += 1
                axis_selection.resize(tuple(new_shape))
                axis_selection[-1, ...] = new_selections[axis]
                selection_refs[axis_index] = axis_selection.shape[0] - 1
            else:
                pass  # Nothing to be done
        self.selection_indexes.resize((self.selection_indexes.shape[0]+1, self.selection_indexes.shape[1]))
        self.selection_indexes[-1, :] = selection_refs

        # Append the description
        self.descriptions.resize((self.descriptions.shape[0]+1,))
        self.descriptions[-1, ...] = annotation.description

        # Update the annotation type
        anno_type = annotation.annotation_type
        current_types = self.annotation_types[:]
        existing_type = anno_type in current_types
        if existing_type:
            type_index = np.where(current_types == anno_type)[0][0]
        else:
            self.annotation_types.resize((self.annotation_types.shape[0]+1,))
            self.annotation_types[-1] = anno_type
            type_index = self.annotation_types.shape[0] - 1
        self.annotation_type_indexes.resize((self.annotation_type_indexes.shape[0]+1, ))
        self.annotation_type_indexes[-1] = type_index


class AnnotationDataGroup(ManagedGroup, AnnotationCollection):
    """
    Managed group for storage of annotations.

    :param hdf_object: The h5py.Group object managed by this class.

    HDF5 file structure:
    -> 'data_object' --> Link to the HDF5 dataset/group we select from
    -> 'annotation_types' --> 1D Dataset with the list of all available types (vlen str)
    -> 'annotation_type_indexes' --> 1D Dataset with the index into the annotation_types array
    -> 'descriptions' --> 1D Dataset with the description of the annotations
    -> 'selection_indexes' --> (#selections, #axis+1) dataset, indicating for each axis the \
        selection that applies or -1 if none. The axes are ordered as [-1, 0, 1, 2, ...]
    -> 'selections_axis_#' --> 2D dataset per each axis with all 1D selections. + (#axis+1) \
        dataset for global selections
    -> axis_index : 1D dataset for the dimension-scale for selection_indexes
    -> collection_description: Attribute with string describing the collection

    """

    # TODO Use the axis_index dataset rather then the implicit axis construction
    # TODO Document the layout of the annotations

    def __init__(self, hdf_object):
        # 1) Call super constructor and initialize the self.hdf_object
        ManagedGroup.__init__(self,
                              hdf_object=hdf_object)
        if not isinstance(hdf_object, h5py.Group):
            raise ValueError("Expected h5py.Group object to initialize AnnotationDataGroup object.")
        if len(hdf_object.keys()) > 0:  # I.e., the group has already been populated
            # 2) Get HDF5 dataset or managed object the annotations apply to
            dataset_spec = self.get_format_specification()['datasets']

            anno_data_object = self.hdf_object[dataset_spec['data_object']['dataset']]
            try:
                managed_object = ManagedObject.get_managed_object(anno_data_object)
            except NameError:
                managed_object = None
            # 2.1) If the data object is a managed object that supports shape and slicing, then use the managed object
            if managed_object is not None:
                if hasattr(managed_object, 'shape') and hasattr(managed_object, '__getitem__'):
                    anno_data_object = managed_object
            # 3) Get array of the indexes  annotation types. 1D int dataset.
            anno_type_indexes = self.hdf_object[dataset_spec['annotation_type_indexes']['dataset']]
            # 4) Get the array of annotation types
            anno_types = self.hdf_object[dataset_spec['annotation_types']['dataset']]
            # 5) Get the array of annotation descriptions. 1D vlen string dataset.
            descriptions = self.hdf_object[dataset_spec['descriptions']['dataset']]
            # 6) 2D dataset of (#selections, #axis+1) describing for each annotation the selections that apply
            selection_refs = self.hdf_object[dataset_spec['selection_indexes']['dataset']]
            # 7) Dict of 2D datasets of per-axis 1D selections + one  (#axis+1) dataset for global selection
            axis_list = [-1] + [index for index in range(len(anno_data_object.shape))]
            data_selections = {}  # The h5py datasets are represented as selections_axis_#
            for axis in axis_list:
                selection_dataset_name = dataset_spec['selections']['prefix'] + str(axis)
                data_selections[axis] = self.hdf_object[selection_dataset_name]
            collection_description = self.hdf_object.attrs['collection_description']
        else:  # The group is empty and will be initialized next via the populate function
            anno_data_object = None
            anno_type_indexes = None
            anno_types = None
            descriptions = None
            selection_refs = None
            data_selections = None
            collection_description = ''
        # Initialize the data annotation collection
        AnnotationCollection.__init__(self,
                                      data=anno_data_object,
                                      annotation_type_indexes=anno_type_indexes,
                                      annotation_types=anno_types,
                                      descriptions=descriptions,
                                      selection_refs=selection_refs,
                                      selections=data_selections,
                                      collection_description=collection_description)

    def __getitem__(self, item):
        """
        If item is a string, then retrieve the corresponding object from the HDF5 file using the
        `ManagedGroup.__getitem__(..)` method, otherwise retrieve the corresponding AnnotationCollection
        using the implementation of `AnnotationCollection.__getitem__(..)`
        """
        if isinstance(item, basestring):
            return ManagedGroup.__getitem__(self, item)
        else:
            return AnnotationCollection.__getitem__(self, item)

    @classmethod
    def get_format_specification(cls):
        """
        Return dictionary describing the specification of the format.
        """
        #raise NotImplementedError('The get_format_specification method must be implemented by the child class')
        return {'datasets': {'data_object': {'dataset': 'data_object',
                                             'prefix': None,
                                             'optional': False,
                                             'dimensions': [],
                                             'description': None,  # We don't use a description as this is a hard-link
                                             'attributes': []},
                             'annotation_types': {'dataset': 'annotation_types',
                                                  'prefix': None,
                                                  'optional': False,
                                                  'dimensions': [{'name': 'annotation_index',
                                                                  'unit': None,
                                                                  'optional': False,
                                                                  'dataset': None,
                                                                  'axis': 0}],
                                                  'dimensions_fixed': True,
                                                  'description': 'List of all available annotation types',
                                                  'attributes': []},
                             'annotation_type_indexes': {'dataset': 'annotation_type_indexes',
                                                         'prefix': None,
                                                         'optional': False,
                                                         'dimensions': [{'name': 'type_index',
                                                                         'unit': None,
                                                                         'optional': False,
                                                                         'dataset': None,
                                                                         'axis': 0}],
                                                         'dimensions_fixed': True,
                                                         'description': "Dataset indicating for each selection the " +
                                                                        "index of the annotation type used. The " +
                                                                        "annotation types are given in the annotation" +
                                                                        " types dataset.",
                                                         'attributes': []},
                             'descriptions': {'dataset': 'descriptions',
                                              'prefix': None,
                                              'optional': False,
                                              'dimensions': [{'name': 'annotation_index',
                                                              'unit': None,
                                                              'optional': False,
                                                              'dataset': None,
                                                              'axis': 0}],
                                              'dimensions_fixed': True,
                                              'description': 'Dataset with the annotation descriptions.',
                                              'attributes': []},
                             'selection_indexes': {'dataset': 'selection_indexes',
                                                   'prefix': None,
                                                   'optional': False,
                                                   'dimensions': [{'name': 'annotation_index',
                                                                   'unit': None,
                                                                   'optional': False,
                                                                   'dataset': None,
                                                                   'axis': 0},
                                                                  {'name': 'axis_index',
                                                                   'unit': 'index',
                                                                   'optional': False,
                                                                   'dataset': 'axis_index',
                                                                   'axis': 1}],
                                                   'dimensions_fixed': True,
                                                   'description': "Dataset indicating for each axis the index of the " +
                                                                  "selection applied to the given axis. -1 indicates " +
                                                                  "that no selection is applied along that axis. The " +
                                                                  "axis index ranges from -1 to n where -1 indicated " +
                                                                  "global selection and n is the number of axes.",
                                                   'attributes': []},
                             'selections': {'dataset': None,
                                            'prefix': 'selections_axis_',
                                            'optional': False,
                                            'dimensions': [{'name': 'selection_index',
                                                            'unit': None,
                                                            'optional': False,
                                                            'dataset': None,
                                                            'axis': 0}],
                                            'dimensions_fixed': False,
                                            'description': "Datasets with all selections for the indicated axis. " +
                                                           "Axis -1 indicates a global selection across all axes. " +
                                                           "One dataset per axis and one for global selection " +
                                                           "(-1) is mandatory.",
                                            'attributes': [{'attribute': 'axis',
                                                            'value': None,
                                                            'prefix': None,
                                                            'optional': False}]}},
                'groups': {},
                'managed_objects': [],
                'attributes': [{'attribute': 'collection_description',
                                'value': None,
                                'prefix': None,
                                'optional': False}],
                'group': None,
                'prefix': "annotations_",
                'optional': True,
                'description': "Managed group for storage of a collection of annotations. Multiple annotation " +
                               "collections may typically be associated with the same data object."}

    def populate(self,
                 data_object,
                 annotation_types=None,
                 annotation_type_indexes=None,
                 descriptions=None,
                 selection_indexes=None,
                 selections=None,
                 collection_description='undefined',
                 annotation_collection=None):
        """
        The populate method is called by the create method after the basic common setup is complete.
        The function should be used to populate the managed object (e.g., add dimensions to a datadset or
        add required datasets to a group. The populate method is passed the kwargs handed to the create
        method.

        :param data_object: The h5py.Dataset (or managed object with support for shape and slicing)
                            the Annotations refer to. (Mandatory)
        :param annotation_types: 1D numpy array of strings (or python list of strings, or h5py dataset of string)
                            with the different possible types of annotations in this collection. May be
                            None in case that AnnotationCollection is initialized as empty.
        :param annotation_type_indexes: 1D h5py.Dataset or numpy array indicating for each annotation the index
                            of the annotation_type that applies.
        :param descriptions: h5py.Dataset or numpy array with the annotation descriptions
        :param selection_indexes: h5py.Dataset or numpy array with the indexes of the selections that apply
        :param collection_description: String with a description of the annotation collection
        :param selections: Python dict of numpy arrays or h5py.Dataset objects. The keys are ints indicating
                           the axis, which must be in the range of [-1,0,1, ... #axis]. Missing values are
                           interpreted as empty (i.e., no selections are available for those axes).
        :param annotation_collection: This parameter may be given instead of the other AnnotationDataGroup
                           specific parameters, i.e., data_object, annotation_types, annotation_type_indexes,
                           descriptions, selection_indexes, selections, collection_description. If given then
                           those parameter will be initialized from the annotation collection instead.
        :type annotation_collection: AnnotationCollection



        :raises: ValueError is raised in case that the object cannot be populated.
        """
        if annotation_collection is not None:
            # Check if any of the other input parameters is given and indicate error in this case
            if data_object is not None or \
                    annotation_types is not None or \
                    annotation_type_indexes is not None or \
                    descriptions is not None or \
                    selection_indexes is not None or \
                    selections is not None or \
                    collection_description != '':
                raise ValueError('AnnotationCollection given for save collides with other input parameters.')
            data_object = annotation_collection.data_object
            annotation_types = annotation_collection.annotation_types
            annotation_type_indexes = annotation_collection.annotation_type_indexes
            descriptions = annotation_collection.descriptions
            selection_indexes = annotation_collection.selection_indexes
            selections = annotation_collection.selections
            collection_description = annotation_collection.collection_description

        # 1) Get the format specification
        format_spec = self.get_format_specification()
        dataset_spec = format_spec['datasets']
        # 1.1) Determine the string type to use based on the available h5py version
        try:
            str_type = h5py.special_dtype(vlen=unicode)
        except NotImplementedError:
            str_type = h5py.special_dtype(vlen=str)

        # 2) Initialize all datasets
        # 2.1) Initialize the data_object dataset
        # 2.1.1) Determine the h5py.Dataset/Group object we point to
        if isinstance(data_object, h5py.Dataset):
            data_hdf5 = data_object
        elif isinstance(ManagedObject, data_object):
            if hasattr(data_object, 'shape') and hasattr(data_object, '__getitem__'):
                data_hdf5 = data_object.get_h5py()
            else:
                raise ValueError("The given managed object does not support data shape  and/or object slicing.")
        else:
            raise ValueError("The given data object does not support annotation (data_object variable error).")
        num_axes = len(data_hdf5.shape)
        axis_list = np.asarray([-1] + [i for i in range(num_axes)], dtype='int32')

        # 2.1.2) Set hard link to the object we annotate
        self.hdf_object[dataset_spec['data_object']['dataset']] = data_hdf5
        self.data_object = data_hdf5  # Initialize the inherited data_object variable of the AnnotationCollection

        # 2.2) Initialize the annotation_types dataset as well as the instance variable of the AnnotationCollection
        if annotation_types is not None:
            if isinstance(annotation_types, list):
                annotation_types_shape = (len(annotation_types),)
            else:
                annotation_types_shape = (annotation_types.shape[0],)
        else:
            annotation_types_shape = (0,)
        # 2.2.1) Create the dataset
        self.annotation_types = self.hdf_object.create_dataset(name=dataset_spec['annotation_types']['dataset'],
                                                               shape=annotation_types_shape,
                                                               dtype=str_type,
                                                               maxshape=(None,),
                                                               chunks=True)
        # 2.2.2) Save the data
        if annotation_types is not None:
            self.annotation_types[:] = annotation_types[:]
        # 2.2.3) Create dimension label
        anno_dimspec_0 = dataset_spec['annotation_types']['dimensions'][0]
        self.annotation_types.dims[anno_dimspec_0['axis']].label = anno_dimspec_0['name']
        # 2.2.4) Add the description attribute
        self.annotation_types.attrs[ManagedObject.description_attribute_name] = \
            dataset_spec['annotation_types']['description']

        # 2.3) Initialize the 'annotation_type_indexes dataset and instance variable from AnnotationCollection
        if annotation_type_indexes is not None:
            annotation_type_indexes_shape = (annotation_type_indexes.shape[0],)
        else:
            annotation_type_indexes_shape = (0,)
        # 2.3.1) Create the dataset
        self.annotation_type_indexes = self.hdf_object.create_dataset(
            name=dataset_spec['annotation_type_indexes']['dataset'],
            shape=annotation_type_indexes_shape,
            dtype='uint32',
            maxshape=(None,),
            chunks=True,
            fillvalue=0)
        # 2.3.2) Save the data
        if annotation_type_indexes is not None:
            self.annotation_type_indexes[:] = annotation_type_indexes[:]
        # 2.3.3) Create the dimension label
        anno_type_dimspec_0 = dataset_spec['annotation_type_indexes']['dimensions'][0]
        self.annotation_type_indexes.dims[anno_type_dimspec_0['axis']].label = anno_type_dimspec_0['name']
        # 2.3.4) Create the description attribute
        self.annotation_type_indexes.attrs[ManagedObject.description_attribute_name] = \
            dataset_spec['annotation_type_indexes']['description']

        # 2.4) Initialize the 'descriptions' dataset and corresponding variable inherited from AnnotationCollections
        if descriptions is not None:
            if isinstance(descriptions, list):
                descriptions_shape = (len(descriptions),)
            else:
                descriptions_shape = (descriptions.shape[0],)
        else:
            descriptions_shape = (0,)
        # 2.4.1) Create the dataset
        self.descriptions = self.hdf_object.create_dataset(
            name=dataset_spec['descriptions']['dataset'],
            shape=descriptions_shape,
            dtype=str_type,
            maxshape=(None,),
            chunks=True)
        # 2.4.2) Save the data
        if descriptions is not None:
            self.descriptions[:] = descriptions[:]
        # 2.4.3) Create the dimension label
        descriptions_dimspec_0 = dataset_spec['descriptions']['dimensions'][0]
        self.descriptions.dims[descriptions_dimspec_0['axis']].label = descriptions_dimspec_0['name']
        # 2.4.4) Create the description attributes
        self.descriptions.attrs[ManagedObject.description_attribute_name] = dataset_spec['descriptions']['description']

        # 2.5) Initialize the 'selection_indexes' dataset and corresponding variable from AnnotationCollections
        if selection_indexes is not None:
            selection_indexes_shape = selection_indexes.shape
            if selection_indexes.shape[1] != (num_axes+1):
                raise ValueError('Mismatch of shape of selection_indexes with number of dimensions of data_object.')
        else:
            selection_indexes_shape = (0, (num_axes+1))
        # 2.5.1) Create the dataset
        self.selection_indexes = self.hdf_object.create_dataset(
            name=dataset_spec['selection_indexes']['dataset'],
            shape=selection_indexes_shape,
            dtype='int32',
            maxshape=(None, selection_indexes_shape[1]),
            compression='gzip',
            compression_opts=1,
            chunks=True,
            shuffle=True,
            fillvalue=-1)
        # 2.5.2) Save the data
        if selection_indexes is not None:
            self.selection_indexes[:] = selection_indexes[:]
        # 2.5.3) Create the dimension label and axis dataset
        sel_index_dimspec = dataset_spec['selection_indexes']['dimensions']
        self.selection_indexes.dims[sel_index_dimspec[0]['axis']].label = sel_index_dimspec[0]['name']
        self.selection_indexes.dims[sel_index_dimspec[1]['axis']].label = sel_index_dimspec[1]['name']
        selection_indexes_axis1_data = axis_list
        selection_indexes_axis1 = self.hdf_object.create_dataset(name=sel_index_dimspec[1]['dataset'],
                                                                 data=selection_indexes_axis1_data)
        self.selection_indexes.dims.create_scale(selection_indexes_axis1, sel_index_dimspec[1]['unit'])
        self.selection_indexes.dims[sel_index_dimspec[1]['axis']].attach_scale(selection_indexes_axis1)
        # 2.5.4) Create the description attributes
        self.selection_indexes.attrs[ManagedObject.description_attribute_name] = \
            dataset_spec['selection_indexes']['description']

        # 2.6) Initialize the datasets for storing all data selections and the corresponding selections variable
        self.selections = {}
        if selections is None:
            selections = {}
        for axis in axis_list:
            # Determine the dataset shape
            if axis in selections:
                selection_shape = selections[axis].shape
            else:
                if axis >= 0:
                    selection_shape = (0, self.data_object.shape[axis])
                else:
                    selection_shape = (0,) + self.data_object.shape
            # Determine the maxhshape
            if axis >= 0:
                selection_maxshape = (None, self.data_object.shape[axis])
            else:
                selection_maxshape = (None,) + self.data_object.shape
            # Determine the data type
            selection_dtype = 'bool'
            # Determine the chunking to be used
            if axis >= 0:
                sfrac = selection_shape[1] / 32768
                if sfrac > 0:
                    selection_chunking = (1, selection_shape[1]/sfrac)
                else:
                    selection_chunking = (int(32768./selection_shape[1] + 0.5), selection_shape[1])
            else:
                selection_chunking = True   # Let h5py handle the chunking for multi-dimensional selections

            # 2.6.1) Create the dataset
            selection_dataset_name = dataset_spec['selections']['prefix'] + str(axis)
            self.selections[axis] = self.hdf_object.create_dataset(
                name=selection_dataset_name,
                shape=selection_shape,
                dtype=selection_dtype,
                maxshape=selection_maxshape,
                compression='gzip',
                compression_opts=2,
                chunks=selection_chunking,
                shuffle=True,
                fillvalue=0)
            # 2.6.2) Save the selection data
            if axis in selections:
                self.selections[axis][:] = selections[axis][:]
            # 2.6.3) Create the dimension labels
            selection_dim_spec = dataset_spec['selections']['dimensions'][0]
            self.selections[axis].dims[selection_dim_spec['axis']].label = selection_dim_spec['name']
            # 2.6.4) Create the descriptions attributes
            self.selections[axis].attrs[ManagedObject.description_attribute_name] = \
                dataset_spec['selections']['description']
            # 2.6.5) Add the axis attribute
            self.selections[axis].attrs[dataset_spec['selections']['attributes'][0]['attribute']] = axis

        # 2.7 Add the collections attributes
        self.hdf_object.attrs[format_spec['attributes'][0]['attribute']] = collection_description
        self.collection_description = collection_description

